""" Module with Sweep class """
from __future__ import annotations
from typing import TYPE_CHECKING
import warnings
import logging

import numpy as np
from qm import qua
from qcodes.parameters import Parameter

from .sequence_parameter import SequenceParameter
if TYPE_CHECKING:
    from .measurement import Measurement

class Sweep:
    """ Class characterizing a parameter sweep along one axis in the OPX """

    _config_to_register = None
    _parameters = None
    _length = None
    _inputs_are_streamed = None
    _input_streams = None
    _can_be_parameterized = None
    snake_scan = False # Assume by default non snake scanning

    def __init__(self, measurement : Measurement, param_dict: dict, register_all = False):
        """
        Constructor class of Sweep class

        Args:
            paret_sequence (SequenceBase): Parent sequence of the sweep
            param_dict (dict): Dict with parameters as keys and arrays as
                setpoints for sweep. If snake_scan is present in the dict,
                use it.
            register_all (bool): Whether all parameters should be registered in
                the QCoDeS measurement
        """
        self.measurement = measurement
        self.register_all = register_all
        self._config = param_dict
        if 'snake' in self._config: # check if the user defined the snake state
            self.snake_scan = self._config['snake']
            del self._config['snake']
        self.configure_sweep()
        self._check_if_parametrizable()

    @property
    def parameters(self):
        """ List containing all varied parameters """
        return self._parameters

    @property
    def qua_variables(self):
        """ Tuple containing all qua variables of parameters """
        return tuple(par.qua_var for par in self.parameters)

    @property
    def qua_sweep_arrays(self):
        """ Tuple containing all qua sweep arrays of parameters """
        return tuple(par.qua_sweep_arr for par in self.parameters)

    @property
    def length(self):
        """ Number of devices for parameters on the given axis """
        return self._length

    @property
    def input_streams(self):
        """Returns all input streams if the sweep is set up to be streamed"""
        return tuple(par.input_stream for par in self.parameters)

    @property
    def inputs_are_streamed(self):
        """Whether sweep is fed by input stream"""
        return self._inputs_are_streamed

    @property
    def can_be_parameterized(self):
        """
        Whether sweep can be parameterized with start, stop and step for memory
        saving. If the user has not set this value, it will be checked by the
        entries of the given arrays
        """
        if self._can_be_parameterized is None:
            return self._check_if_parametrizable()
        else:
            return self._can_be_parameterized

    @can_be_parameterized.setter
    def can_be_parameterized(self, value: bool):
        """Setter for can_be_parameterized"""
        if value is False:
            self._can_be_parameterized = False
            for parameter in self.parameters:
                parameter.can_be_parameterized = False
        elif value is True:
            if self._can_be_parameterized is True:
                self._can_be_parameterized = True
        else:
            raise ValueError("can_be_parameterized must be of type bool")

    @property
    def config(self) -> dict:
        """Config dict for parameter sweep. Keys are params, values setpoints"""
        return self._config

    @property
    def config_to_register(self) -> list:
        """ Parameters that will be registered in QCoDeS measurement """
        return self._config_to_register

    @config_to_register.setter
    def config_to_register(self, param_list: list) -> None:
        """Setter for config_to_register"""
        if all(param in self.parameters for param in param_list):
            self._config_to_register = param_list
        else:
            raise KeyError(
                "Some of the given parameters are not within the swept params")

    def configure_sweep(self) -> None:
        """Configures the sweep from the given dictionairy"""
        self.check_input_dict()
        self._parameters = []
        self._config_to_register = {}
        for i, parameter in enumerate(self.config.keys()):
            self._parameters.append(parameter)
            ### Remove scalar validators and setup the sweep_validator
            while parameter.remove_validator():
                pass
            parameter.add_validator(parameter.sweep_validator)
            parameter.validate(self.config[parameter])
            ### Usually not all params are registered (issue with live plotting)
            if self.register_all:
                setpoints = parameter.convert_to_real_units(
                    self.config[parameter])
                self._config_to_register[parameter] = setpoints*parameter.scale
                parameter.set(setpoints)
            elif i == 0:
                value = self.config[parameter]
                setpoints = parameter.convert_to_real_units(self.config[parameter])
                if isinstance(value, (list, np.ndarray)):
                    if parameter.scale is None:
                        raise ValueError(
                            f"Parameter scale of '{parameter.full_name}' is None."
                            " Must be set for sweep arrays. Check the type of"
                            " the parameter in the config file. If it does not"
                            " have a type, set the scale manually."
                            )
                    parameter.set(setpoints)
                    self._config_to_register[parameter] = setpoints*parameter.scale
                elif isinstance(value, int):
                    ### creates a mock set of values (stream array indices)
                    self._config_to_register[parameter] = np.arange(value)
                else:
                    raise ValueError(
                        "Keys in sweep dictionairies must be of type int, list"
                        f"or numpy.ndarray, is:  {type(value)}")
            if isinstance(self.config[parameter], int):
                parameter.input_stream = True
                #parameter.add_stream_param_to_sequence()
        if all(param.input_stream is not None for param in self.parameters):
            self._inputs_are_streamed = True
        else:
            self._inputs_are_streamed = False

    def check_input_dict(self) -> None:
        """
        Validates equal sizes of input arrays in three steps:
            1) Checks if all parameters are SequenceParameter/Parameter
            2) Checks if var_type is int, bool or qua.fixed
            3) Checks if all sweep setpoint arrays have same lengths
            4) Checks if all input streams have the same dimension
        """
        self._config = self._convert_str_keys_to_param(self._config)
        for param in self.config.keys():
            if not isinstance(param, (SequenceParameter, Parameter)):
                raise TypeError(
                    f"Key {param} in sweep config must be of type "
                    "SequenceParameter or Parameter")
            if not param.var_type in (int, bool, qua.fixed):
                raise TypeError(
                    f"Key {param.full_name} in sweep config must have a var_type"
                    f" of int, bool, or qua.fixed. Is: {param.var_type}."
                )
        param_arrays = []
        param_streams = []
        for values in self._config.values():
            if isinstance(values, np.ndarray):
                param_arrays.append(values)
            elif isinstance(values, int):
                param_streams.append(values)
            else:
                raise ValueError(
                        "Keys in sweep dictionairies must be of type int, list"
                        f" or numpy.ndarray, is:  {type(values)}")
        ### checks if all setpoint arrays have the same length
        if param_arrays:
            list_iter = iter(param_arrays)
            length = len(next(list_iter))
            if not all(len(l) == length for l in list_iter):
                raise ValueError('not all lists have same length!')
            self._length = length
        # checks if all input streams have the same length and compares to
        # regular sweeps
        if param_streams:
            if not all(size == param_streams[0] for size in param_streams):
                raise ValueError(
                    f"not all param streams have same size! {param_streams}")
            if param_arrays:
                if self._length != param_streams[0]:
                    raise ValueError(
                        "When mixing stream variable sweeps with static sweeps,"
                        f" size of input stream ({param_streams[0]})  must be"
                        f" same as static sweep array ({self._length})"
                    )
            self._length = param_streams[0]

    def _convert_str_keys_to_param(self, param_dict: dict) -> dict:
        """
        Converts string keys to SequenceParameter objects

        Args:
            param_dict (dict): Dict with string keys

        Returns:
            dict: Dict with SequenceParameter keys
        """
        new_dict = {}
        for key, value in param_dict.items():
            if isinstance(key, str):
                new_key = self.measurement.find_parameter_from_str_path(key)
                new_dict[new_key] = value
            else:
                new_dict[key] = value
        return new_dict

    def _check_if_parametrizable(self) -> bool:
        """
        Checks whether the sweep array can be defined in terms of start, step
        and stop (memory saving).

        Returns:
            bool: Whether the sweep can be parametrized
        """
        if self.inputs_are_streamed:
            return False
        parameterizability_list = []
        for param in self.parameters:
            sweep_arr = self.config[param]
            sweep_arr_differences = np.ediff1d(sweep_arr)
            mean_step = np.mean(sweep_arr_differences)
            if np.std(sweep_arr_differences) > np.abs(0.1*mean_step):
                can_be_parameterized = False
            else:
                can_be_parameterized = True
            param.can_be_parameterized = can_be_parameterized
            parameterizability_list.append(can_be_parameterized)
        ### In case not all parameters can be parametrized, none can
        parameterizabel = any(parameterizability_list)
        if not parameterizabel:
            for param in self.parameters:
                param.can_be_parameterized = False
        return parameterizabel

    def qua_generate_parameter_sweep(self, next_action: callable, next_sweep) -> None:
        """
        Runs a qua loop based on the configured method. Currently three
        different methods are available:
            1) From input stream
            2) From parametrized array (start, stop, step)
            3) From an explicitly defined qua array
        If necessary, we need to reset the snake var for the inner loop before
        entering the snaking loop.
        """

        # Reset the snake var if necessary
        if next_sweep is not None:
            ns_sv = next_sweep.get_snake_var()
            if ns_sv is not None:
                qua.assign(next_sweep.get_snake_var(), True)

        if self.inputs_are_streamed:
            self._qua_input_stream_loop(next_action)
        elif self.can_be_parameterized:
            self._qua_parmetrized_loop(next_action)
        else:
            self._qua_explicit_array_loop(next_action)

    def get_snake_var(self):
        """
        Declare a snake_var if we are snaking and return that var if snaking.

        Returns:
            qua bool var : The snaking indicator if snaking, None otherwise
        """
        if self.snake_scan:
            if not hasattr(self, 'sweep_snake_var'):
                self.sweep_snake_var = qua.declare(bool, True)
            return self.sweep_snake_var
        return None

    def _qua_input_stream_loop(self, next_action: callable) -> None:
        """Runs a qua for loop for an array that is imported from a stream"""
        warnings.warn("Input streaming is not fully supported")
        for param in self.parameters:
            qua.wait(int(1e6))
            qua.advance_input_stream(param.input_stream)
            logging.debug(
                "Assigning %s with length %s (input stream)",
                param.name, self.length)

        sweep_idx_var = qua.declare(int)
        qua.assign(sweep_idx_var, 0)
        with qua.while_(sweep_idx_var < self.length):
            for param in self.parameters:
                qua.assign(
                    param.qua_var, param.input_stream[sweep_idx_var])
            next_action()
            self._advance_step_counter(sweep_idx_var)

    def _qua_parmetrized_loop(self, next_action: callable) -> None:
        """
        Runs a qua for loop from parametrized qua_arange. Start, stop and step
        are calculated from the input array

        Args:
            next_action (callable): Next action to be executed after the loop
        """
        ### Define start, stop and step for all params that can be parameterized
        parameters_sss = self._parameterize_sweep()

        ### Declaring sweep index variable and respective loop for sweep
        sweep_idx_var = qua.declare(int)
        if self.snake_scan:
            qua.assign(self.get_snake_var(), ~self.get_snake_var())

        qua.assign(sweep_idx_var, 0)
        with qua.while_(sweep_idx_var < self.length):
            for param in self.parameters:
                if not param.can_be_parameterized:
                    qua.assign(
                        param.qua_var, param.qua_sweep_arr[sweep_idx_var])
            if not self.snake_scan:
                for param, sss in parameters_sss.items():
                    self._qua_calc_param_step(param, sss, sweep_idx_var, False)
            else:
                with qua.if_(self.get_snake_var()):
                    for param, sss in parameters_sss.items():
                        self._qua_calc_param_step(param, sss, sweep_idx_var, True)
                with qua.else_():
                    for param, sss in parameters_sss.items():
                        self._qua_calc_param_step(param, sss, sweep_idx_var, False)

            qua.align()

            ### This is where either the whole sequence or the next sweep is run
            next_action()
            self._advance_step_counter(sweep_idx_var)

    def _qua_calc_param_step(self, param, sss, sweep_idx_var, reverse):
        """
        Calculates the step within a parameter sweep.
        
        Args:
            param (SequenceParameter): Parameter to be swept
            sss (dict): Dict with start, stop and step (parameterizing sweep)
            sweep_idx_var (qua variable): Index variable for sweep
            reverse (bool): Whether the sweep is reversed (from stop to start)

        Raises:
            TypeError: If var_type is not int or fixed

        Returns:
            None
        """
        ### Note this implementation is written in a very explicit way avoiding
        ### multiplications to save lines of code. This is done to avoid
        ### multiplications in the FPGA code (e.g (-1)*x)
        if param.var_type == int:
            if not reverse:
                qua.assign(
                    param.qua_var, sss['start'] + sss['step']*sweep_idx_var)
            else:
                qua.assign(
                    param.qua_var, sss['stop'] - sss['step']*sweep_idx_var)
        elif param.var_type == qua.fixed:
            if not reverse:
                qua.assign(
                    param.qua_var,
                    sss['start'] + qua.lib.Cast.mul_fixed_by_int(
                        sss['step'], sweep_idx_var)
                    )
            else:
                qua.assign(
                    param.qua_var,
                    sss['stop'] - qua.lib.Cast.mul_fixed_by_int(
                        sss['step'],sweep_idx_var)
                    )
        else:
            raise TypeError(
                "Only int and fixed qua types are supported for param sweeps"
                )

    def _qua_explicit_array_loop(self, next_action):
        """Runs a qua for loop from explicitly defined qua arrays"""
        for param in self.parameters:
            logging.debug(
                "Assigning %s to %s (loop)",
                    param.name, param.qua_sweep_arr)

        sweep_idx_var = qua.declare(int)
        qua.assign(sweep_idx_var, 0)
        with qua.while_(sweep_idx_var < self.length):
            for param in self.parameters:
                qua.assign(
                    param.qua_var, param.qua_sweep_arr[sweep_idx_var])
            next_action()
            self._advance_step_counter(sweep_idx_var)


    def _parameterize_sweep_array(
        self, param: SequenceParameter, sweep_array: np.ndarray
        ) -> tuple[int, int, int] | tuple[float, float, float]:
        """
        Parameterizes start, stop and step from given sweep array

        Args:
            param (SequenceParameter): Sweep-parameter
            sweep_array (np.ndarray): Array to be parameterized

        Returns:
            (int, int, int) | (float, float, float): Start, stop and step
        """
        start = sweep_array[0]
        stop = sweep_array[-1]
        step = np.mean(np.ediff1d(sweep_array))
        step *= 0.999
        if param.var_type == int:
            start, stop = int(sweep_array[0]), int(sweep_array[-1] + step)
            step = round(step)

        ### This can probably done more elegantly
        length_of_array = len(np.arange(start, stop, step))
        while length_of_array != self.length:
            if length_of_array > self.length:
                stop -= step
            elif length_of_array < self.length:
                stop += step
            length_of_array = len(np.arange(start, stop, step))

        if not length_of_array == self.length:
            raise ValueError(
                "Sweep array must have same length as sweep length or one more."
                f"Is {length_of_array}, should be {self.length}"
                )
        return start, stop, step

    def _parameterize_sweep(self):
        """
        Parameterizes the sweep array in start, stop, step for all paramerters
        with sweep arrays with equal step size
        """
        parameters_sss = {}
        for param in self.parameters:
            if param.can_be_parameterized:
                start, stop, step = self._parameterize_sweep_array(
                    param, self.config[param]*param.scale)
                parameters_sss[param] = {
                    'start': start,
                    'stop': stop,
                    'step': step
                    }
                length_of_array = len(np.arange(start, stop, step))
                warnings.warn(
                    f"\n\tYour input array of length {length_of_array} "
                    f"for {param.name} will be parametrized with\n\t"
                    f"start {start}, step {step}, stop {stop}"
                    f" of length {length_of_array}. \n\tCheck output!",
                    category=ResourceWarning
                    )
        return parameters_sss

    def _advance_step_counter(self, sweep_idx_var):
        """
        Advances the sweep index variable by one step and checks if the
        step requirements are met. This is ONLY doen for the inner most sweep
        hence this will automatically stop advancement of the all outer sweeps

        Args:
            sweep_idx_var (qua variable): Sweep index variable to be advanced
        """
        def step_counter():
            qua.assign(sweep_idx_var, sweep_idx_var + 1)
        if self.measurement.sweeps[-1] == self:
            self.measurement.qua_check_step_requirements(step_counter)
        else:
            step_counter()


"""Module containing generic sequence for a simple square pulse sequence"""
from qm import qua
from arbok_driver import SubSequence, Sample, qua_helpers

class SquarePulse2(SubSequence):
    """
    Class containing parameters and sequence for a simple square pulse
    """

    def qua_sequence(self):
        """Macro that will be played within the qua.program() context"""
        qua.align(*self.sticky_elements())
        qua_helpers.arbok_go(
                sub_sequence = self,
                elements= self.sticky_elements(),
                from_volt = 'vHome',
                to_volt = 'vSquare',
                duration = self.ramp_time(),
                operation = 'unit_ramp',
                )
        qua.wait(self.t_square_pulse(), *self.sticky_elements())
        qua_helpers.arbok_go(
                sub_sequence = self,
                elements= self.sticky_elements(),
                from_volt = 'vSquare',
                to_volt = 'vHome',
                duration = self.ramp_time(),
                operation = 'unit_ramp',
                )


"""Module testing the Program class"""
import pytest
from arbok_driver import Measurement, ArbokDriver, Device

def test_arbok_driver_init(arbok_driver)-> None:
    """Tests if arbok_driver is correctly initialized"""
    assert arbok_driver.name == 'arbok_driver'
    assert arbok_driver.device.name == 'dummy_device'
    assert arbok_driver.measurements == []

def test_arbok_driver_adding_measurement(arbok_driver, dummy_device)-> None:
    """Tests if measurement is correctly added to arbok_driver"""
    meas = Measurement(arbok_driver, 'dummy_measurement', dummy_device)
    assert len(arbok_driver.measurements) == 1
    assert arbok_driver.measurements[0] == meas
    assert len(arbok_driver.submodules) == 1

    arbok_driver.reset_measurements()
    assert len(arbok_driver.measurements) == 0
    assert len(arbok_driver.submodules) == 0



import pytest
import tempfile

from arbok_driver import SubSequence
from arbok_driver.parameter_types import Voltage, Time

import logging
logging.basicConfig(
    filename = 'logs/sequence_params_test.txt',
    filemode = 'w',
    encoding = 'utf-8',
    level = logging.DEBUG
)

square_conf2 = {
    'sticky_elements': {
        'value': ['gate_1', 'gate_2', 'gate_3', 'gate_4'],
        'unit': 'gate label'
    },
    'vHome': {
        'type' : Voltage,
        "label": 'Default voltage point during the sequence',
        'elements': {
            'gate_1': 0,
            'gate_2': 0,
            'gate_3': 0,
            'gate_4': 0,
        }
    },
    'vSquare': {
        'type' : Voltage,
        "label": 'Voltage amplitude of square pulse',
        'elements': {
            'gate_1': 0.1,
            'gate_2': -0.05,
            'gate_3': 0.08,
            'gate_4': 0.25,
        }
    },
    't_square_pulse': {
        'value': 100,
        'type': Time
    },
    'ramp_time': {
        'value': 20,
        'type': Time
    },
}


expected_output = """square_conf2:
	parameter      value
--------------------------------------------------------------------------------
IDN             :	None 
ramp_time       :	20 (cycles)
sticky_elements :	['gate_1', 'gate_2', 'gate_3', 'gate_4'] (gate label)
t_square_pulse  :	100 (cycles)
vHome_gate_1    :	0 (V)
vHome_gate_2    :	0 (V)
vHome_gate_3    :	0 (V)
vHome_gate_4    :	0 (V)
vSquare_gate_1  :	0.1 (V)
vSquare_gate_2  :	-0.05 (V)
vSquare_gate_3  :	0.08 (V)
vSquare_gate_4  :	0.25 (V)"""

def compare_multiline_strings(string1, string2):
    lines1 = string1.strip().splitlines()
    lines2 = string2.strip().splitlines()

    # Ensure both strings have the same number of lines
    assert len(lines1) == len(lines2), "Number of lines mismatch"

    for line_num, (line1, line2) in enumerate(zip(lines1, lines2), start=1):
        assert line1 == line2, f"Line {line_num} differs:\n'{line1}'\nvs\n'{line2}'"

def test_parameters(dummy_device, capfd) -> None:
    sequenceBase = SubSequence('square_conf2', dummy_device, square_conf2)
    sequenceBase.print_readable_snapshot()
    out, err = capfd.readouterr()
    compare_multiline_strings(out, expected_output)



import pytest
import copy

import numpy as np
from qcodes.instrument import Instrument

from arbok_driver import Sweep
from arbok_driver import SequenceParameter




@pytest.fixture
def mock_instrument() -> Instrument:
    dummy_instrument = Instrument("dummy_instrument")
    dummy_instrument.add_parameter(
        parameter_class = SequenceParameter,
        name  = 'Par1',
        element = "P1",
        config_name = 'dummy',
    )
    dummy_instrument.add_parameter(
        parameter_class = SequenceParameter,
        name  = 'Par2',
        element = "P1",
        config_name = 'dummy',
    )
    yield dummy_instrument
    dummy_instrument.__del__()

@pytest.fixture
def sweep_dict(mock_instrument) -> dict:
    sweep_dict = {
        mock_instrument.Par1: np.ones((10)),
        mock_instrument.Par2: np.zeros((10))
        }
    return sweep_dict

@pytest.fixture
def sweep_dict_with_array_mismatch(sweep_dict, mock_instrument) -> dict:
    sweep_dict_new = copy.copy(sweep_dict)
    sweep_dict_new[mock_instrument.Par1] = np.zeros((11))
    return sweep_dict_new

@pytest.fixture
def sweep_dict_with_wrong_key(sweep_dict) -> dict:
    sweep_dict_new = copy.copy(sweep_dict)
    sweep_dict_new['no_param'] = np.ones((10))
    return sweep_dict_new

def test_sweep_instanciation(
        sweep_dict,
        sweep_dict_with_array_mismatch,
        sweep_dict_with_wrong_key) -> None:
    sweep = Sweep(sweep_dict)
    assert sweep.length == 10
    assert len(sweep.parameters) == 2
    assert len(sweep.config_to_register) == 1
    sweep = Sweep(sweep_dict, register_all = True)
    assert len(sweep.config_to_register) == 2
    with pytest.raises(KeyError):
        sweep.config_to_register = ['no_param']
    with pytest.raises(AttributeError):
        sweep.length = 4
    with pytest.raises(ValueError):
        _ = Sweep(sweep_dict_with_array_mismatch)
    with pytest.raises(TypeError):
        _ = Sweep(sweep_dict_with_wrong_key)
    



#############
# VARIABLES #
#############
qop_ip = "127.0.0.1"
cluster_name = "my_cluster"

readout_len = 100
readout_amp = 0.001

time_of_flight = 300
hold_offset_duration = 200

gate_1_amp = 0.25
gate_2_amp = 0.25

opx_config = {
    "version": 1,
    "controllers": {
        "con1": {
            "analog_outputs": {
                1: {"offset": 0.0},
                2: {"offset": 0.0},
                3: {"offset": 0.0},
                4: {"offset": 0.0},
                5: {"offset": 0.0},
            },
            "analog_inputs": {
                1: {"offset": 0.0, "gain_db": 0},
                2: {"offset": 0.0, "gain_db": 0},
            },
        },
    },
    "elements": {
        "gate_1": {
            "singleInput": {
                "port": ("con1", 1),
            },
            "hold_offset": {"duration": hold_offset_duration},  # in clock cycles (4ns)
            "operations": {
                "bias": "bias_gate_1_pulse",
            },
        },
        "gate_2": {
            "singleInput": {
                "port": ("con1", 2),
            },
            "hold_offset": {"duration": hold_offset_duration},  # in clock cycles (4ns)
            "operations": {
                "bias": "bias_gate_2_pulse",
            },
        },
        "gate_3": {
            "singleInput": {
                "port": ("con1", 3),
            },
            "hold_offset": {"duration": hold_offset_duration},  # in clock cycles (4ns)
            "operations": {
                "bias": "bias_gate_1_pulse",
            },
        },
        "gate_4": {
            "singleInput": {
                "port": ("con1", 4),
            },
            "hold_offset": {"duration": hold_offset_duration},  # in clock cycles (4ns)
            "operations": {
                "bias": "bias_gate_2_pulse",
            },
        },
        "readout_element": {
            "singleInput": {
                "port": ("con1", 5),
            },
            "operations": {
                "readout": "readout_pulse",
            },
            "intermediate_frequency": 0,
            "outputs": {
                "out1": ("con1", 1),
                "out2": ("con1", 2),
            },
            "time_of_flight": time_of_flight,
            "smearing": 0,
        },
    },
    "pulses": {
        "bias_gate_1_pulse": {
            "operation": "control",
            "length": 16,
            "waveforms": {
                "single": "bias_gate_1_pulse_wf",
            },
        },
        "bias_gate_2_pulse": {
            "operation": "control",
            "length": 16,
            "waveforms": {
                "single": "bias_gate_2_pulse_wf",
            },
        },
        "readout_pulse": {
            "operation": "measurement",
            "length": readout_len,
            "waveforms": {
                "single": "readout_pulse_wf",
            },
            "integration_weights": {
                "cos": "cosine_weights",
                "sin": "sine_weights",
                "minus_sin": "minus_sine_weights",
            },
            "digital_marker": "ON",
        },
    },
    "waveforms": {
        "bias_gate_1_pulse_wf": {"type": "constant", "sample": gate_1_amp},
        "bias_gate_2_pulse_wf": {"type": "constant", "sample": gate_2_amp},
        "readout_pulse_wf": {"type": "constant", "sample": readout_amp},
        "zero_wf": {"type": "constant", "sample": 0},
    },
    "digital_waveforms": {
        "ON": {"samples": [(1, 0)]},
    },
    "integration_weights": {
        "cosine_weights": {
            "cosine": [(1.0, readout_len)],
            "sine": [(0.0, readout_len)],
        },
        "sine_weights": {
            "cosine": [(0.0, readout_len)],
            "sine": [(1.0, readout_len)],
        },
        "minus_sine_weights": {
            "cosine": [(0.0, readout_len)],
            "sine": [(-1.0, readout_len)],
        },
    },
}


import copy
import re
import pytest
import numpy as np
from qm import generate_qua_script

from arbok_driver import SubSequence, Device, SequenceParameter
from arbok_driver.tests.helpers import set_sweeps_args

def test_sub_seq_param_init(sub_sequence_1, sub_sequence_2) -> None:
    """Tests if parameters are correctly initialized from config"""
    assert sub_sequence_1.par1.get() == 1
    assert sub_sequence_2.par4.get() == 2
    assert len(sub_sequence_1.parameters) == 6
    assert len(sub_sequence_2.parameters) == 5
    assert isinstance(sub_sequence_1.par1.get(), int)

def test_parent_child_sequence_behaviour(
        parent_sub_sequence, sub_sequence_1) -> None:
    """Tests the behaviour of nested sub sequences"""
    assert parent_sub_sequence.sub_seq1.par2.get() == 10
    assert parent_sub_sequence.sub_seq2.vHome_P1.get() == 0
    assert parent_sub_sequence.sub_seq1.measurement == parent_sub_sequence
    with pytest.raises(KeyError):
        parent_sub_sequence.add_subsequence(sub_sequence_1)
        print(parent_sub_sequence.submodules)

def test_qua_program_compilation_wo_sweeps(
        dummy_measurement) -> None:
    """Tests if any sub sequence within a sequence that has sweeps compiles 
    correctly to QUA code considering those sweeps """
    dummy_measurement.set_sweeps(*set_sweeps_args(dummy_measurement))
    qua_prog_str = dummy_measurement.dummy_parent_sub.get_qua_program_as_str()
    # we expect 8 declares: 2x3(2 per parameter -> sweep_arr + qua_var) 
    # + 2 as iterators for for loops (per sweep axis)
    assert len([m.start() for m in re.finditer('declare', qua_prog_str)]) == 8
    # we expect 3, one per parameter in total
    assert len([m.start() for m in re.finditer('assign', qua_prog_str)]) == 3
    # we expect 2 one per sweep axis
    assert len([m.start() for m in re.finditer('for_', qua_prog_str)]) == 2
    #breakpoint()
    qua_prog_str = dummy_measurement.dummy_parent_sub.sub_seq1.get_qua_program_as_str()
    assert len([m.start() for m in re.finditer('declare', qua_prog_str)]) == 8
    assert len([m.start() for m in re.finditer('assign', qua_prog_str)]) == 3
    assert len([m.start() for m in re.finditer('for_', qua_prog_str)]) == 2


def test_add_qc_params_from_config(sub_sequence_1) -> None:
    """Test adding qc params by config"""
    sub_sequence_1.add_qc_params_from_config({
        'par8': {'unit': 'cycles', 'value': int(8)},
        'vRead': {'unit': 'V', 'elements': {'P1': 0.3, 'J1': 7, }},
        })
    assert isinstance(sub_sequence_1.par8, SequenceParameter)
    assert sub_sequence_1.par8() == 8
    assert sub_sequence_1.vRead_J1() == 7
    with pytest.raises(KeyError):
        sub_sequence_1.add_qc_params_from_config({
            'par69': {'unit': 'cycles', 'toast': int(8)},
            })

def test_find_measurement(mock_program, dummy_measurement) -> None:
    """Tests if parent sequence is found from sub sequences"""
    mock_program.add_measurement(dummy_measurement)
    parent = mock_program.dummy_measurement.measurement
    assert parent == mock_program.dummy_measurement
    parent = mock_program.dummy_measurement.dummy_parent_sub.measurement
    assert parent == mock_program.dummy_measurement


# #!/usr/bin/env python
# import os
# import tempfile

# from arbok_driver import ArbokDriver, Device, ReadSequence
# from arbok_driver.measurement import Measurement
# from types import SimpleNamespace

# from all_the_qua import AllTheQua
# from arbok_driver.parameter_types import Voltage
# import qcodes as qc
# from qm_config.opx1000 import config

# # get some parameters which we can sweep
# sequence_config = {
#     'parameters': {
#         'v_set_home': {
#             'type': Voltage,
#             'elements': { 'output': 0, 'loopback': 0 }
#         },
#     },
# }

# device = SimpleNamespace(**{'elements' : [ 'output', 'loopback' ], 'divider_config' : {} })
# device.reload_master_config = lambda: None  # Method that does nothing but pass
# device.config = config
# arbok_driver = ArbokDriver('arbok_driver', device)

# measurement = Measurement(
#     parent = arbok_driver,
#     name = 'measurement_name',
#     device = device,
#     sequence_config = sequence_config
# )
# atq = AllTheQua(measurement, 'atq', device)

# measurement.qc_measurement_name = 'sweep_measurement'
# measurement.qc_experiment = qc.dataset.load_or_create_experiment('experiment.name', 'device.name')

# import numpy as np
# v_start = 0.1
# v_range = np.linspace(-v_start, v_start, 100)
# v_arr = np.array([0, 1, 10, 100], dtype=float)

# # sweeps = [{'v_set_home_P1': v_range, 'snake': True},
# #         {'v_set_home_J1': v_arr},
# #         {'v_set_home_P2': 32, 'snake': False},
# #         ]
# sweeps = [
#     {'v_set_home_output': v_range, 'snake': False},
#         ]

# measurement.set_sweeps(*sweeps)
# # measurement.register_gettables(keywords=[('atq_stream')])

# tmp_dir = tempfile.gettempdir()
# file_path = os.path.join(tmp_dir, "arbok_test", "sweep_param_test.qua")
# os.makedirs(os.path.dirname(file_path), exist_ok=True)

# measurement.print_qua_program_to_file(file_path)

# sweep_list_arg = [{arbok_driver.iteration: np.arange(1)}]
# run_loop = measurement.get_measurement_loop_function(sweep_list_arg)

# # arbok_driver.connect_opx(host_ip = '192.168.88.254', port = 9510, log_level = 'DEBUG')
# # measurement.compile_qua_and_run()

# # try:
# #     dataset = run_loop()
# # except KeyboardInterrupt:
# #     print('Dumping execution report:')
# #     arbok_driver.qm_job.execution_report()



""" Module containing Devices class """
import warnings
from .utils import get_module

class Device():
    """
    Class describing the used device by its config and the used sequence. 
    """

    _master_config_path = None
    master_config = None

    def __init__(
            self, name: str,
            opx_config: dict,
            divider_config: dict,
            master_config = None,
            ):
        """
        Constructor class for 'Device' class.

        Args:
            name (str): Name of the used device
            opx_config (dict): Configuration dictionary for the OPX
            divider_config (dict): Configuration dictionary for the divider
            master_config (dict): Configuration dictionary for universally accessible
                parameters across all sub-sequences as well as default
                sequence configurations.
        """
        self.name = name
        self.config = opx_config
        self.elements = list(self.config['elements'].keys())
        self.divider_config = divider_config
        self._update_master_config(master_config)

    def _update_master_config(self, master_config = None):
        """
        Update the master config from a dict

        Args:
            master_config (dict): Configuration dictionary for universally accessible
                parameters across all sub-sequences as well as default
                sequence configurations.
        """
        if master_config is None:
            master_config = {}
        if 'parameters' in master_config:
            self.param_config = master_config['parameters']
        else:
            self.param_config = {}
        if 'default_sequence_configs' in master_config:
            self.default_sequence_configs = master_config['default_sequence_configs']
        else:
            self.default_sequence_configs = {}

    @property
    def master_config_path(self):
        """
        Getter for master_config.

        Returns:
            str: The current configuration dictionary.
        """
        return self._master_config_path

    @master_config_path.setter
    def master_config_path(self, config_path):
        """
        Setter for master_config_path.
        Validates and sets the configuration path as a string.

        Args:
            config_path (str): A dictionary containing configuration data.

        Raises:
            ValueError: If the config is not a string.
        """
        if not isinstance(config_path, str):
            raise ValueError("master_config_path must be a str.")
        self._master_config_path = config_path
        mc = get_module('mc', config_path)
        if not hasattr(mc, 'config'):
            raise AttributeError(
                "Dictionary 'config' not found in the file: "
                f"{self._master_config_path}"
                )
        self._update_master_config(mc.config)

    def reload_master_config(self):
        """
        If the master config path is set, force reloading
        """
        if self._master_config_path is not None:
            mcp = self.master_config_path
            self.master_config_path = mcp



"""Module containing threshold readout helper"""
import logging
from qm import qua
from arbok_driver import ReadSequence, AbstractReadout

class VarReadout(AbstractReadout):
    """Helper class enabling variable readout from a given signal"""
    signals: list

    def __init__(
            self,
            name: str,
            sequence: ReadSequence,
            attr_name: str,
            save_results: bool,
            params: dict
            ):
        """
        Constructor method for `VarReadout` readout helper class
        
        Args:
            name (str): Name of readout
            sequence (arbok_driver.ReadSequence): ReadSequence performing the
                abstract readout
            attr_name (str): Name under which the result will be saved
            save_results (bool): Whether to save the results
            params (dict): Parameters to be added to the read sequence
        """
        super().__init__(name, sequence, attr_name, save_results, params)
        self._create_gettables()

    def qua_measure(self):
        """Measures the given gettables and assigns the result to the vars"""
        for signal in self.signal_names:
            qua.assign(
                self.variable_obs[signal].qua_var, self.variables[signal]()
                )

    def _create_gettables(self):
        """Creates the gettables for the given readout"""
        # TODO: write


"""Module containing configuration for eight dot device"""
from arbok_driver.parameter_types import (
    Time, Voltage, Frequency, Amplitude, List, Int, Radian
)
from qcodes import validators

master_config = {
    'parameters': {
        'iteration': {
            'type': Int,
            'value': int(1),
            'vals': validators.Ints(),
        },
        'gate_elements': {
            'type': List,
            'value': [
                'P1', 'J1',
                'P2', 'J2',
                'P7', 'J7',
                'P8',
            ]
        },
        'readout_elements': {
            'type': List,
            'value': [
                'SET1',
                'SET2',
            ]
        },
        'qubit_elements': {
            'type': List,
            'value': [
                'Q1', 'Q2',
                # 'Q3', 'Q4',
                # 'Q5', 'Q6',
                # 'Q7', 'Q8',
                'qe1','qe2',
                ]
        },
        'f_larmor': {
            'type': Frequency,
            'label': 'resonance frequency',
            'elements': {
                'Q1': int(27.85e6),
                'Q2': int(30.1e6),
            },
        },
    },
    'default_sequence_configs': {
    },
}





divider_config = {
    'gate_1': {
        'division': 1,
    },
    'gate_2': {
        'division': 2,
    },
    'gate_2': {
        'division': 2.5,
    },
    'readout_element': {
        'division': 1
    }
}




# Single QUA script generated at 2025-09-03 16:51:02.305902
# QUA library version: 1.2.1

from qm import CompilerOptionArguments
from qm.qua import *

with program() as prog:
    v1 = declare(int, value=0)
    with infinite_loop_():
        pause()
        assign(v1, 0)
        align()
        play("ramp"*amp(0.1), "gate_1", duration=200)
        wait(100, "gate_1")
        play("ramp"*amp(-0.1), "gate_1", duration=200)
        align()
        assign(v1, (v1+1))
        r0 = declare_stream()
        save(v1, r0)
        align()
    with stream_processing():
        r0.buffer(1).save("mock_driver_mock_measurement_shots")


config = None

loaded_config = None




# #!/usr/bin/env python
# import os
# import tempfile

# from arbok_driver import ArbokDriver, Device, ReadSequence
# from arbok_driver.measurement import Measurement
# from types import SimpleNamespace

# from all_the_qua import AllTheQua
# from arbok_driver.parameter_types import Voltage
# import qcodes as qc
# from qm_config.opx1000 import config

# # get some parameters which we can sweep
# measurement_config = {
#     'parameters': {
#         'v_set_home': {
#             'type': Voltage,
#             'elements': { 'element_a': 0, 'element_b': 0, 'element_c': 0, 'element_d': 0 }
#         },
#     },
# }

# device = SimpleNamespace(**{'elements' : [ 'element_a', 'element_b', 'element_c', 'element_d' ], 'divider_config' : {} })
# device.config = config
# arbok_driver = ArbokDriver('arbok_driver', device)

# measurement = Measurement(
#     parent = arbok_driver,
#     name = 'measurement_name',
#     device = device,
#     sequence_config = measurement_config
# )

# from arbok_driver.tests.var_readout import VarReadout
# available_abstract_readouts = {
#     'var_readout': VarReadout
# }
# from arbok_driver.tests.var_readout_config import var_readout_config
# atq = AllTheQua(measurement, 'atq', device, var_readout_config, available_abstract_readouts = available_abstract_readouts)

# measurement.qc_measurement_name = 'sweep_measurement'
# measurement.qc_experiment = qc.dataset.load_or_create_experiment('experiment.name', 'device.name')

# import numpy as np
# v_start = 0.1
# v_range_a = np.linspace(-v_start, v_start, 10)
# v_range_b = np.linspace(-v_start, v_start, 20)
# v_range_c = np.linspace(-v_start, v_start, 30)
# v_arr = np.array([0, 1, 10, 100], dtype=float)

# sweeps = [
#     {'v_set_home_element_a': v_range_a},
#     {'v_set_home_element_b': v_range_b},
#     {'v_set_home_element_c': v_range_c,
#      'v_set_home_element_d': v_range_c, 'snake': True},
#         ]

# measurement.set_sweeps(*sweeps)
# measurement.register_gettables(atq.gettables[0])

# tmp_dir = tempfile.gettempdir()
# file_path = os.path.join(tmp_dir, "arbok_test", "sweep_param_snake_test.py")
# os.makedirs(os.path.dirname(file_path), exist_ok=True)
# measurement.print_qua_program_to_file(file_path)

# sweep_list_arg = [{arbok_driver.iteration: np.arange(1)}]
# run_loop = measurement.get_measurement_loop_function(sweep_list_arg)

# arbok_driver.connect_opx(host_ip = '192.168.88.254', port = 9510, log_level = 'DEBUG')
# measurement.compile_qua_and_run()

# try:
#     dataset = run_loop()
# except KeyboardInterrupt:
#     print('Dumping execution report:')
#     arbok_driver.qm_job.execution_report()


"""Module containing generic sequence for a simple square pulse sequence"""
from qm import qua
from arbok_driver import SubSequence

class SquarePulse(SubSequence):
    """
    Class containing parameters and sequence for a simple square pulse
    """

    def qua_sequence(self):
        """Macro that will be played within the qua.program() context"""
        qua.align()
        qua.play('ramp'*qua.amp(self.amplitude()), self.element(), duration = self.t_ramp())
        qua.wait(self.t_square_pulse(), self.element())
        qua.play('ramp'*qua.amp(-self.amplitude()), self.element(), duration = self.t_ramp())




"""Module containing the ReadSequence class"""
from __future__ import annotations
from typing import TYPE_CHECKING
import logging
from qcodes.validators import Arrays

from .abstract_readout import AbstractReadout
from .gettable_parameter import GettableParameter
from .signal import Signal
from .sub_sequence import SubSequence

if TYPE_CHECKING:
    from .device import Device

class ReadSequence(SubSequence):
    """ Baseclass for sequences containing readouts """
    def __init__(
        self,
        parent: SubSequence,
        name: str,
        device: Device,
        sequence_config: dict,
        ):
        """
        Constructor class for ReadSequence class
        Args:
            parent (SubSequence): Parent instrument module
            name (dict): name of the ReadSequence
            device (Device): device for which sequence is configured
            sequence_config (dict): Dict configuring all parameters for the given
                read-sequence and its read-points and abstract-readouts
        """
        super().__init__(
            parent=parent,
            name=name,
            device=device,
            sequence_config=sequence_config
            )

        self._check_read_sequence_config(sequence_config)
        self.sequence_config = sequence_config
        self._signals = {}
        self._readout_groups = {}
        self._abstract_readouts = {}

        logging.debug("Adding signals to ReadSequence: %s", self.name)
        self._add_signals_from_config(self.sequence_config['signals'])
        logging.debug("Adding readout groups to ReadSequence: %s", self.name)
        self._add_readout_groups_from_config(self.sequence_config["readout_groups"])
        self.measurement.add_available_gettables(self._gettables)

    @property
    def signals(self):
        """Signals registered in this sequence"""
        return self._signals

    @property
    def readout_groups(self):
        """All configured readout groups"""
        return self._readout_groups

    @property
    def abstract_readouts(self):
        """All abstract readouts in this read-sequence"""
        return self._abstract_readouts

    def qua_declare(self):
        """
        QUA variable and stream declaration based on the given sequence
        configuration. Only to be called within qua.program() context manager!
        """
        for readout_name, abstract_readout in self._abstract_readouts.items():
            logging.debug("Declaring qua vars for %s", readout_name)
            abstract_readout.qua_declare_variables()

    def qua_stream(self):
        """
        Saves acquired results to qua stream
        Only to be called within qua.program() context manager!
        """
        for readout_name, abstract_readout in self._abstract_readouts.items():
            logging.debug("Saving streams of abstract readout %s", readout_name)
            abstract_readout.qua_save_streams()
            continue

    def _add_signals_from_config(self, signal_name_list: list):
        """
        Creates all signals on the ReadSequence from the given config.
        Checks validity of signal names and adds them to the sequence.
        Signals are added as attributes to the sequence with the same name.
        Signals are also added to the signals property of the sequence.

        Args:
            signal_name_list (list): List of signal names to be added to the

        Raises:
            TypeError: If signal name is not a list
            ValueError: If signal name is not a string or already exists
            AttributeError: If signal name is already an attribute of the sequence
        """
        if not isinstance(signal_name_list, list):
            raise ValueError(
                f"Signals in sequence config of {self.full_name} must be a list!"
            )
        for signal_name in signal_name_list:
            logging.debug("Adding signal %s to %s", signal_name, self.full_name)
            if not isinstance(signal_name, str):
                raise ValueError(
                    f"Signal name {signal_name} in sequence config of"
                    f" {self.full_name} must be a string!"
                )
            if hasattr(self, signal_name):
                raise AttributeError(
                    f"Attribute '{signal_name}' already exists on {self.full_name}!"
                    " Cant add signal with the same name."
                )
            new_signal = Signal(name=signal_name, read_sequence = self)
            self._signals[new_signal.name] = new_signal
            setattr(self, signal_name, new_signal)

    def _check_read_sequence_config(self, sequence_config: dict):
        """
        Checks the given sequence config for validity
        Args:
            sequence_config (dict): Dict configuring all parameters for the given
                read-sequence and its read-points and abstract-readouts
        """
        if 'signals' not in sequence_config:
            raise ValueError(
                f"No signals configured in sequence config for {self.full_name}!"
                )
        if 'readout_groups' not in sequence_config:
            raise ValueError(
                "No readout groups configured in sequence config for"
                f"{self.full_name}!"
            )
        group_configs = sequence_config['readout_groups'].values()
        if not all([isinstance(conf, dict) for conf in group_configs]):
            raise TypeError(
                f"Readout groups configs in sequence config of {self.full_name}"
                " must be dicts!"
            )

    def _add_readout_groups_from_config(self, groups_config: dict) -> None:
        """
        Creates readout groups from config file

        Args:
            groups_config (dict): Dict containing readout groups and their
                readouts. The keys are the group names and the values are dicts
                containing readout names and their configurations.
        Raises:
            TypeError: If readout config is not a dict
        """
        for group_name, group_conf in groups_config.items():
            logging.debug("Creating readout group %s", group_name)
            self._readout_groups[group_name] = {}
            for readout_name, readout_conf in group_conf.items():
                if not isinstance(readout_conf, dict):
                    raise TypeError(
                        f"Readout config of {readout_name} in group {group_name}"
                        f" of {self.full_name} must be a dict!"
                    )
                abstract_readout = self._add_abstract_readout(
                    readout_conf, readout_name, group_name
                )
                self._readout_groups[group_name][readout_name] = abstract_readout

    def _add_abstract_readout(
        self,
        readout_conf: dict,
        readout_name: str,
        group_name: str
    ) -> AbstractReadout:
        """
        Adds an abstract readout to the sequence from the given configuration.
        
        Args:
            readout_conf (dict): Configuration dict of the readout
            readout_name (str): Name of the readout
            group_name (str): Name of the readout group
        Raises:
            ValueError: If signal is not found in the sequence signals
        Returns:
            AbstractReadout: The created abstract readout object
        """
        full_readout_name = f"{group_name}__{readout_name}"
        self._check_readout_config(readout_conf, full_readout_name, group_name)
        logging.debug(
            "Adding '%s' readout to sequence '%s' with name '%s'",
            readout_conf['readout_class'].__name__,
            self.full_name,
            full_readout_name,
        )
        if "save_results" in readout_conf:
            save_results = readout_conf["save_results"]
        else:
            save_results = True
        kwargs = readout_conf['kwargs'] if 'kwargs' in readout_conf else {}
        params = readout_conf["parameters"] if "parameters" in readout_conf else None
        if readout_conf['signal'] not in self.signals:
            raise ValueError(
                f"Signal {readout_conf['signal']} not found in read sequence"
                f"{self.full_name}. Available signals: {list(self.signals.keys())}"
            )
        abstract_readout = readout_conf['readout_class'](
            name = full_readout_name,
            read_sequence = self,
            signal = self.signals[readout_conf['signal']],
            save_results = save_results,
            parameters = params,
            **kwargs
        )
        self._abstract_readouts[full_readout_name] = abstract_readout
        return abstract_readout

    def _check_readout_config(
            self, readout_conf: dict, readout_name: str, group_name: str) -> None:
        """
        Checks the given readout configuration for validity.
        
        Args:
            readout_conf (dict): Configuration dict of the readout
            readout_name (str): Name of the readout
            group_name (str): Name of the readout group
        Raises:
            KeyError: If 'readout_class' or 'signal' is not in the config
            TypeError: If 'readout_class' is not a subclass of AbstractReadout
        """
        if 'readout_class' not in readout_conf:
            raise KeyError(
                f"'readout_class' not found in readout config for {readout_name}"
                f" in group {group_name} on {self.full_name}!."
                f" Config is: {readout_conf}"
            )
        if 'signal' not in readout_conf:
            raise KeyError(
                f"'signal' not found in readout config for {readout_name}"
                f" in group {group_name} on {self.full_name}!"
                )
        if not issubclass(readout_conf['readout_class'], AbstractReadout):
            raise TypeError(
                f"Readout class {readout_conf['readout_class'].__name__} is not"
                f" a subclass of AbstractReadout ({self.full_name},"
                f" {group_name}__{readout_name})!"
            )

    def add_gettable(self, gettable: GettableParameter) -> None:
        """
        Adds a gettable parameter to the read sequence. This is used to make
        the gettable fetchable during a measurement.

        Args:
            gettable (GettableParameter): Gettable parameter to be added
        """
        if not isinstance(gettable, GettableParameter):
            raise TypeError(
                f"Expected GettableParameter, got {type(gettable)}"
            )
        self._gettables.append(gettable)

    def get_qm_elements_from_signals(self):
        """
        Gets all qm elements from the configured signals in this read sequence
        """
        qm_elements = []
        for _, signal in self.signals.items():
            qm_elements += signal.readout_elements.values()
        # this is meant to be deterministic with non duplicates in the list.
        return list(dict.fromkeys(qm_elements))



from typing import Union, Optional
import logging
import math

from qm import qua

from arbok_driver import SubSequence, SequenceParameter

def arbok_go(
        sub_sequence: SubSequence,
        elements: list,
        to_volt: str | list,
        operation: str,
        from_volt: Optional[Union[str, list]] = None,
        duration: any = None,
        align_after: str = 'elements',
        no_play_tolerance: float = 1e-6
    ):
    """ 
    Helper function that `play`s a qua operation on the respective elements 
    specified in the sequence config.
    TODO:   - [ ] raise error when target and origin dims dont match
            - [ ] raise error if duration is too short
            - [ ] separate this functions into digestible parts

    Args:
        sub_sequence (Sequence): Sequence from which parameters are fetched
        elements (list): elements on which pulse is applied
        to_volt (str, List): voltage point to move to
        operation (str): Operation to be played -> find in OPX config
        from_volt (str, List): voltage point to come from
        duration (str | qcodes.Parameter): duration of the operation 
        align_after (optional, bool, str): whether to align channels after
            ramping. Input 'all' alligns all. 'none' does not align 
            at all and 'elements' (default) aligns all given elements
        no_play_tolerance (optional, float): tolerance for not playing a pulse
    """
    if isinstance(duration, SequenceParameter):
        duration = duration()

    from_volt = _check_voltage_point_input(from_volt)
    to_volt = _check_voltage_point_input(to_volt)
    from_volt_signs = {point: 1 for point in from_volt}
    to_volt_signs = {point: 1 for point in to_volt}

    for i, point in enumerate(from_volt):
        if point[0] == '-':
            from_volt[i] = point[1:]
            from_volt_signs[point[1:]] = -1
    for i, point in enumerate(to_volt):
        if point[0] == '-':
            to_volt[i] = point[1:]
            to_volt_signs[point[1:]] = -1
    origin_param_sets = sub_sequence.find_parameters_from_keywords(
        from_volt,
        elements
        )
    target_param_sets = sub_sequence.find_parameters_from_keywords(
        to_volt,
        elements
        )
    if not set(origin_param_sets.keys()) == set(target_param_sets.keys()):
        raise KeyError(
            "All given voltage points must have the same elements in the conf",
            f"{elements}")
    _qua_align(
        method=align_after,
        elements = elements,
        sub_sequence = sub_sequence
        )
    for element in elements:
        sticky_pulse_amplitude = _calculate_sticky_pulse_amplitude(
            target_param_sets[element],
            origin_param_sets[element],
            to_volt_signs,
            from_volt_signs
        )
        kwargs = {
            'pulse': operation*qua.amp(sticky_pulse_amplitude),
            'element': element
            }
        if duration is not None:
            kwargs['duration'] = duration

        logging.debug(
            "Arbok_go: Moving %s from %s to %s by %s in %s", 
            element, from_volt, to_volt, sticky_pulse_amplitude, sub_sequence
            )
        if not isinstance(sticky_pulse_amplitude, float):
            qua.play(**kwargs)
        elif math.isclose(sticky_pulse_amplitude, 0, abs_tol= no_play_tolerance):
            logging.debug(
                "Arbok_go: Omitting %s since amplitude %s is small (th = %s)", 
                element, sticky_pulse_amplitude, no_play_tolerance
                )
        else:
            qua.play(**kwargs)

    _qua_align(
        method=align_after,
        elements = elements,
        sub_sequence = sub_sequence
        )

def _calculate_sticky_pulse_amplitude(
        target_points: dict,
        origin_points: dict,
        to_volt_signs: dict,
        from_volt_signs: dict
        ) -> float:
    """
    Calculates the amplitude of the sticky pulse to be played. Target and origin
    can have multiple points. The amplitude is calculated as the difference

    Args
        element (str): qua element on which pulse is played
        target_points (dict): dict containing the target points as keys
            and their params (either values or qua variables) as values
        origin_points (dict): dict containing the origin points as keys
            and their params (either values or qua variables) as values
        to_volt_signs (dict): dict containing the target points as keys and
            their signs as values
        from_volt_signs (dict): dict containing the origin points as keys and
            their signs as values
    Returns:
        float | any: amplitude of the sticky pulse in qm units (note real volts)
    """
    target_v, origin_v = 0, 0
    for point, param in target_points.items():
        target_v += param()*(to_volt_signs[point])
    for point, param in origin_points.items():
        origin_v += param()*(from_volt_signs[point])


    return target_v - origin_v

def _qua_align(method, elements, sub_sequence):
    match method:
        case "elements":
            qua.align(*elements)
        case 'none':
            pass
        case 'global':
            qua.align()
        case 'sequence':
            qua.align(*sub_sequence.elements)
        case _:
            raise ValueError(
                "Argument for align can only be bool or 'elements'")

def _check_voltage_point_input(points: list | str) -> list:
    """
    Checks if input is of type list or string and returns input as list
    """
    if points is None:
        return ['vHome']
    elif isinstance(points, str):
        return [points]
    elif isinstance(points, list):
        return points
    else:
        raise ValueError("Must be of type str or list!")

def reset_elements(
        element_list: list,
        sub_sequence: any,
        align_method: str = 'sequence',
        ) -> None:
    """runs `ramp_to_zero` on all element in the given list"""
    for element_name in element_list:
        qua.ramp_to_zero(element_name)
    _qua_align(
        method=align_method,
        elements=element_list,
        sub_sequence=sub_sequence
        )



"""
QUA-Config supporting OPX1000
"""

sampling_rate = int(1e9) # Hz
T = 5e-3 # s

config = {
    "version": 1,
    "controllers": {
        'con1': {
            "type": "opx1000",
            "fems": {
                1: {
                    "type": "LF",
                    "analog_outputs": {
                        1: {
                            "offset": 0.0,
                            "output_mode": "direct",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "mw",
                        },
                        2: {
                            "offset": 0.0,
                            "output_mode": "direct",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "mw",
                        },
                        3: {
                            "offset": 0.0,
                            "output_mode": "direct",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "mw",
                        },                        
                        4: {
                            "offset": 0.0,
                            "output_mode": "direct",
                            "sampling_rate": sampling_rate,
                            "upsampling_mode": "mw",
                        },
                    },
                    "digital_outputs": {},
                    "analog_inputs": {
                        1: {"offset": 0.0, "gain_db": 0, "sampling_rate": sampling_rate},
                        2: {"offset": 0.0, "gain_db": 0, "sampling_rate": sampling_rate},
                    },
                }
            },
        }
    },
    "elements": {
        "element_a": {
            "singleInput": {
                "port": ("con1", 1, 1),
            },
            "operations": {
                "cw": "const_pulse",
            },
        },
        "element_b": {
            "singleInput": {
                "port": ("con1", 1, 2),
            },
            "operations": {
                "readout": "readout_pulse",
            },
            "outputs": {
                "out1": ("con1", 1, 1),
                "out2": ("con1", 1, 2),
            },
            "time_of_flight": 180,
            "smearing": 0,
        },
        "element_c": {
            "singleInput": {
                "port": ("con1", 1, 3),
            },
            "operations": {
                "cw": "const_pulse",
            },
        },
        "element_d": {
            "singleInput": {
                "port": ("con1", 1, 4),
            },
            "operations": {
                "cw": "const_pulse",
            },
        },
    },
    "pulses": {
        "const_pulse": {
            "operation": "control",
            "length": 100,
            "waveforms": {
                "single": "const_wf",
            },
        },
        "readout_pulse": {
            "operation": "measurement",
            "length": int(T*1e9), # ns
            "waveforms": {
                "single": "readout_wf",
            },
            "digital_marker": "ON",
        },
    },
    "waveforms": {
        "const_wf": {"type": "constant", "device": 0.249},
        "readout_wf": {"type": "constant", "device": 0}
    },
    "digital_waveforms": {
        "ON": {"devices": [(1, 0)]},
    },
}



"""Module containing the MeasurementRunner class."""
from __future__ import annotations
from typing import TYPE_CHECKING
import logging
import copy
import time
import warnings

import numpy as np
from rich.progress import Progress

if TYPE_CHECKING:
    from arbok_driver import Measurement

class MeasurementRunner:
    """
    Helper class constructing QCoDeS measurement loops
    """

    def __init__(
        self,
        measurement: Measurement,
        sweep_list: list[dict] | None = None,
        register_all: bool = False
        ):

        self.measurement = measurement
        self.qc_measurement  = measurement.get_qc_measurement()
        self.sweep_list = sweep_list

        if self.sweep_list is not None:
            sweep_lengths = [len(next(iter(dic.values()))) for dic in sweep_list]
            self.nr_total_results =  np.prod(sweep_lengths)
        else:
            self.nr_total_results = 1
            self.sweep_list = []
            warnings.warn(
                "NO sweeps outside the OPX registerd. Thus the measurement only"
                " fills the buffer once and finishes "
                "(e.g only lower progres bar)")

        self.result_args_dict = self._get_result_arguments(register_all)
        self.inner_func = None
        self.progress_tracker = None
        self.progress_bars = None
        self.counter = 0

        self.datasaver = None

    def run_arbok_measurement(self, inner_func: callable = None):
        """
        Runs the measurement with the given inner function.
        
        Args:
            inner_func (callable): The function to be executed for each measurement
                point. It should accept the datasaver and the current sweep values
                as arguments.
        Returns:
            dataset (Dataset): The QCoDeS dataset containing the measurement results.
        """
        logging.debug("Preparing params and gettables for measurement")
        self.inner_func = inner_func
        self._prepare_params_and_gettables()
        # Run the measurement with the recursive measurement loop
        try:
            logging.debug("Running measurement with %s", self.measurement.name)
            self.datasaver = self._build_qc_measurement()
        except KeyboardInterrupt:
            logging.debug("Measurement interrupted by user.")
            print("Measurement interrupted by user.")
            return

    def _build_qc_measurement(self):
        """
        Builds the QCoDeS measurement object for the measurement.
        """
        self.counter = 0
        with self.qc_measurement.run() as datasaver:
            self.datasaver = datasaver
            with Progress() as self.progress_tracker:
                ### Adding progress bars

                self.progress_bars = {}
                total_progress = self.progress_tracker.add_task(
                    description = f"[green]Total progress...\n0/{self.nr_total_results}",
                    total = self.nr_total_results)
                batch_progress = self.progress_tracker.add_task(
                    description = "[cyan]Batch progress...",
                    total = self.measurement.sweep_size)
                self.progress_bars['total_progress'] = total_progress
                self.progress_bars['batch_progress'] = batch_progress
                self._create_recursive_measurement_loop(
                    self.sweep_list, datasaver)
            print("Measurement finished!")
        return datasaver

    def _create_recursive_measurement_loop(
            self,
            sweep_list: list[dict],
            datasaver
            ):
        """
        Creates a recursive measurement loop over the sweep_list.

        Args:
            sweep_list (list[dict]): List of dictionaries containing the sweep
                parameters.
            datasaver (DataSaver): The QCoDeS DataSaver object to save results to.
        """
        # Copy to avoid modifying the original list
        sweep_list = copy.copy(sweep_list)
        if not sweep_list:
            # If the sweep list is empty, execute the inner function
            if self.inner_func is not None:
                self.inner_func()
            ### Program is resumed and all gettables are fetched when ready
            if not self.measurement.driver.is_mock:
                self.measurement.driver.qm_job.resume()
            time.sleep(0.1)
            logging.debug("Job resumed, Fetching gettables")
            self.measurement.wait_until_result_buffer_full(
                (self.progress_bars['batch_progress'], self.progress_tracker)
            )
            self._save_results(datasaver)
            return

        ### The first axis will be popped from the list and iterated over
        sweep_dict = sweep_list.pop(0)
        sweep_axis_size = len(list(sweep_dict.values())[0])
        if sweep_axis_size <= 2:
            warnings.warn(
                "Measurements with axis sizes less of 2 or less are not recommended."
                " Results may not be displayable with plottr."
            )
        for idx in range(sweep_axis_size):
            ### The parameter values are set for the current iteration
            for param, values in sweep_dict.items():
                value = values[idx]
                logging.debug('Setting %s to %s', param.instrument.name, value)
                param.set(value)
                ### The parameter is added to the result arguments dict if its
                ### dict entry is registered
                if param in self.result_args_dict:
                    self.result_args_dict[param] = (param, value)
                else:
                    logging.debug( "Param %s on %s not registered",
                        param.instrument, param.name)
            self._create_recursive_measurement_loop(
                sweep_list = sweep_list,
                datasaver = datasaver,
            )

    def _save_results(self, datasaver):
        """
        Saves the results of the measurement to the datasaver.
        
        Args:
            datasaver (DataSaver): The QCoDeS DataSaver object to save results to.
        """
        self.counter += 1
        result_args_temp = []
        for _, gettable in self.measurement.gettables.items():
            result = gettable.fetch_results()
            result_args_temp.append(
                (gettable, result)
            )
        ### Retreived results are added to the datasaver
        result_args_temp += list(self.result_args_dict.values())
        datasaver.add_result(*result_args_temp)
        title = "[green]Total progress\n "
        self.progress_tracker.update(
            self.progress_bars['total_progress'],
            advance=1,
            description=f"{title}{self.counter}/{self.nr_total_results}"
            )
        self.progress_tracker.refresh()
        logging.debug("Results saved")

    def _get_result_arguments(self, register_all: bool = False) -> dict:
        """
        Generates a dict of parameters that are varied in the sweeps.
        The dict will be used to register the results in the measurement.
        
        Args:
            register_all (bool): If True, all settables will be registered in the
                measurement. If False, only the first settable of each axis will be
                registered
        Returns:
            result_args_dict (dict): Dict with parameters as keys and tuples of
                (parameter, value) as values. The tuples will be used for
                `add_result` in the measurement
        """
        gettable_setpoints = []
        result_args_dict = {}
        for i, sweep_dict in enumerate(self.sweep_list):
            for j, param in enumerate(list(sweep_dict.keys())):
                if j == 0 or register_all:
                    result_args_dict[param] = ()
                    gettable_setpoints.append(param)
                else:
                    logging.debug(
                        "Not adding settable %s on axis %s", param.name, i)
        return result_args_dict

    def _prepare_params_and_gettables(self):
        """
        Prepare parameters and gettables for the measurement.
        """
        for param, _ in self.result_args_dict.items():
            logging.debug(
                "Registering sequence parameter %s", param.full_name)
            self.qc_measurement.register_parameter(param)

        for _, gettable in self.measurement.gettables.items():
            gettable_setpoints = list(self.result_args_dict.keys())
            logging.debug("Registering gettable %s", gettable_setpoints)
            self.qc_measurement.register_parameter(
                gettable, setpoints = gettable_setpoints)



# #!/usr/bin/env python
# from arbok_driver import ArbokDriver, Measurement, SubSequence
# from arbok_driver.parameter_types import Voltage, Time, String
# import numpy as np

# from helpers import qua_program_trim_gen_time, compare_qua_programs

# from qm import qua

# class SquarePulse(SubSequence):
#     """
#     Class containing parameters and sequence for a simple square pulse
#     """

#     def qua_sequence(self):
#         """Macro that will be played within the qua.program() context"""
#         qua.align()
#         qua.play('ramp'*qua.amp(self.amplitude()), self.element(), duration = self.ramp_time())
#         qua.wait(self.t_square_pulse(), self.element())
#         qua.play('ramp'*qua.amp(-self.amplitude()), self.element(), duration = self.ramp_time())

# square_conf = {
#     'amplitude': {
#         'value': 0.5,
#         'type': Voltage,
#     },
#     't_square_pulse': {
#         'value': 100,
#         'type': Time
#     },
#     'element': {
#         'type': String,
#         'value': 'gate_1',
#         'unit': 'gate label'
#     },
#     'ramp_time': {
#         'value': 20,
#         'type': Time
#     },
# }

# from arbok_driver import Device
# from arbok_driver.tests.dummy_opx_config import dummy_qua_config, divider_config

# dummy_device = Device('dummy_device', dummy_qua_config, divider_config)
# qm_driver = ArbokDriver('qm_driver', dummy_device)
# dummy_measurement = Measurement(qm_driver, 'dummy_measurement', dummy_device)

# square_pulse = SquarePulse(dummy_measurement, 'square_pulse', dummy_device, square_conf)

# dummy_measurement.set_sweeps(
#     {
#         square_pulse.amplitude: np.linspace(0.1, 1, 5)
#     }
# )

# # dummy_measurement.reset()
# qua_program = dummy_measurement.get_qua_program_as_str()
# qp1 = qua_program_trim_gen_time(qua_program)

# dummy_measurement.reset()

# qua_program2 = dummy_measurement.get_qua_program_as_str()
# qp2 = qua_program_trim_gen_time(qua_program2)

# line_differences = compare_qua_programs(qp1, qp2)
# if len(line_differences):
#     raise Exception("Programes should have been identical, but weren't")



"""Module containing functions for path finding objects within the Arbok driver."""
from __future__ import annotations
from typing import TYPE_CHECKING
from functools import reduce


if TYPE_CHECKING:
    from .abstract_readout import AbstractReadout
    from .gettable_parameter import GettableParameter
    from .measurement import Measurement
    from .read_sequence import ReadSequence
    from .signal import Signal
    from .sub_sequence import SubSequence

def get_gettable_from_read_sequence(
        read_sequence: ReadSequence,
        signal_name: str,
        gettable_name: str,
        ) -> GettableParameter:
    """
    Helper method to get a gettable from this readout. This is used

    Args:
        read_sequence (ReadSequence): The read sequence to get the gettable from
        signal_name (str): Name of the signal
        gettable_name (str): Name of the gettable
    
    Raises:
        KeyError: If the signal or gettable is not found in the read sequence
        KeyError: If the GettableParameter is not found in the signal
        ValueError: If the found object is not a child class of GettableParameter
    """
    from .gettable_parameter import GettableParameter
    if signal_name not in read_sequence.signals:
        raise KeyError(
            f"Signal {signal_name} not found in read sequence"
            f" '{read_sequence.name}'. Available signals: "
            f"{read_sequence.signals.keys()}"
        )
    signal: Signal = read_sequence.signals[signal_name]
    if gettable_name not in signal.gettables:
        raise KeyError(
            f"Gettable {gettable_name} not found in signal {signal.name}"
            f" of read sequence '{read_sequence.name}'. "
            f"Available gettables: {signal.gettables.keys()}"
        )
    gettable: GettableParameter = signal.gettables[gettable_name]
    if not isinstance(gettable, GettableParameter):
        raise ValueError(
            f"The given path {signal_name}.{gettable_name} yields a ",
            f"{type(gettable)}-type, not a child class of gettable"
        )
    return gettable

def get_sequence_from_measurement_path(
        measurement: Measurement,
        path: str | list[str]
        ) -> SubSequence:
    """
    Get a sequence from a measurement by resolving the given attribute path.

    Args:
        measurement (Measurement): The measurement to get the read sequence from
        path (str): The attribute path to the read sequence relative to the measurement

    Returns:
        SubSequence: The sub-sequence found at the given path
    """
    if isinstance(path, str):
        path: list[str] = path.split('.')
    sequence = reduce(getattr, path, measurement)
    from .sub_sequence import SubSequence
    if not isinstance(sequence, SubSequence):
        raise ValueError(
            f"The given path {path} does not resolve to a SubSequence"
            f" on measurement {measurement.name}. Got {type(sequence)} instead."
        )
    return sequence

def get_read_sequence_from_measurement_path(
        measurement: Measurement,
        path: str | list[str]
) -> ReadSequence:
    """
    Get a read sequence from a measurement by resolving the given attribute path.

    Args:
        measurement (Measurement): The measurement to get the read sequence from
        path (str): The attribute path to the read sequence relative to the measurement

    Returns:
        ReadSequence: The read sequence found at the given path
    """
    if isinstance(path, str):
        path: list[str] = path.split('.')
    sub_sequence = get_sequence_from_measurement_path(measurement, path)
    from .read_sequence import ReadSequence
    if not isinstance(sub_sequence, ReadSequence):
        raise ValueError(
            f"The given path {path} does not resolve to a ReadSequence on "
            f"measurement {measurement.name}. Got {type(sub_sequence)} instead."
        )
    return sub_sequence

def get_read_sequence_from_readout_path(
        readout: AbstractReadout,
        path: str | list[str]
) -> ReadSequence:
    """
    Get a read sequence from a readout by resolving the given attribute path.

    Args:
        readout (AbstractReadout): The readout to get the read sequence from
        path (str): The attribute path to the read sequence relative to the readout

    Returns:
        ReadSequence: The read sequence found at the given path
    """
    measurement = readout.read_sequence.measurement
    return get_read_sequence_from_measurement_path(measurement, path)

def get_params_with_prefix(sub_sequence: SubSequence, prefix: str) -> dict:
    """
    Finds all parameters in a sub-sequence that have the given prefix in their name.
    TODO: This is not a very clean solution, make more readable

    Args:
        prefix (str): Prefix of the element parameters
    
    Returns:
        dict: Dictionary with suffixes as keys and parameters as values
    """
    all_params = sub_sequence.parameters
    full_prefix = f"{sub_sequence.name}__{prefix}"
    param_names = [x for x in all_params if full_prefix in x]
    suffix_list = [x.split(full_prefix)[-1].split('_')[-1] for x in param_names]
    return {e: all_params[p] for e, p in zip(suffix_list, param_names)}

def get_gettable_from_measurement_path(
        measurement: Measurement,
        path: str | list[str]) -> GettableParameter:
    """
    Get a gettable from a measurement by resolving the given attribute path.

    Args:
        measurement (Measurement): The measurement to get the gettable from
        path (str): The attribute path to the gettable relative to the measurement
    
    Returns:
        GettableParameter: The gettable found at the given path
    """
    if isinstance(path, str):
        path: list[str] = path.split('.')

    if len(path) <= 2:
        raise ValueError(
            f"The given path {path} is too short to resolve a gettable. "
            "Please provide a path with the format "
            "'sub_sequence_path.signal.gettable_name'."
        )
    signal_name, gettable_name = path[-2:]
    sub_sequence_path = path[:-2]
    read_sequence = get_read_sequence_from_measurement_path(
        measurement, sub_sequence_path)
    return get_gettable_from_read_sequence(
        read_sequence, signal_name, gettable_name
    )



import pytest 

from arbok_driver import ArbokDriver, Device, SubSequence, Measurement, parameter_types
from arbok_driver.tests.dummy_opx_config import dummy_qua_config

opx_scale = 2
divider_config = {
    'gate_1': {
        'division': 1*opx_scale,
    },
    'gate_2': {
        'division': 1*opx_scale,
    },
    'readout_element': {
        'division': 1*opx_scale
    }
}

@pytest.fixture
def dummy_device():
    """Returns dummy device instance"""
    return Device(
        name = 'dummy_device',
        opx_config = dummy_qua_config,
        divider_config = divider_config,
    )

@pytest.fixture
def arbok_driver(dummy_device):
    """Returns ArbokDriver instance"""
    driver = ArbokDriver(name = 'arbok_driver', device = dummy_device)
    yield driver
    driver.__del__()
    del driver

@pytest.fixture
def dummy_measurement(arbok_driver, dummy_device):
    """Returns dummy measurement instance"""
    measurement = Measurement(arbok_driver, 'dummy_measurement', dummy_device)
    yield measurement
    measurement.__del__()
    del measurement

@pytest.fixture
def sub_sequence_1(dummy_measurement, dummy_device):
    """Returns dummy subsequence with a few parameters"""
    config_1 = {
        'par1': {'type': parameter_types.Int, 'value': int(1)},
        'par2': {'type': parameter_types.Int, 'value': int(10)},
        'par3': {'type': parameter_types.Voltage, 'value': 1.1},
        'vHome': {'type': parameter_types.Voltage, 'elements': {'P1': 5, 'J1': 6, }},
    }
    seq1 = SubSequence(dummy_measurement, 'sub_seq1', dummy_device, config_1)
    yield seq1
    #seq1.__del__()

@pytest.fixture
def sub_sequence_2(dummy_measurement, dummy_device):
    """Returns dummy subsequence with a few parameters"""
    config_2 = {
        'par4': {'type': parameter_types.Int, 'value': int(2)},
        'par5': {'type': parameter_types.Int, 'value': int(3)},
        'vHome': {'type': parameter_types.Voltage, 'elements': {'P1': 0, 'J1': 7, }},
    }
    seq2 = SubSequence(dummy_measurement, 'sub_seq2', dummy_device, config_2)
    yield seq2
    seq2.__del__()
